#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>
#include <stdint.h>
#include "immutable_treap.h"
#include "random.h"
#include <limits.h>
#include <pthread.h>
struct treap_impl {
    int key;
    struct treap_impl * left;
    struct treap_impl * right;
    unsigned int priority;
};

typedef struct treap_impl treap;


/************* Help Structures **************/

struct thread_local_rand_state {
    char pad1[124];
    unsigned int rand;
    char pad2[128];
};


#ifdef USE_THREAD_LOCAL_RANDOM
static __thread struct thread_local_rand_state local_random_state = {.rand = 0};
#endif

static int thread_number = 0;
pthread_mutex_t global_lock = PTHREAD_MUTEX_INITIALIZER;


/************* Help Functions ***************/

/*
 * Creates a random integer that can't be zero
 */
unsigned int random_int(void * p) {
#ifdef NO_RANDOM
  return 42;
#elif defined  USE_THREAD_LOCAL_RANDOM
    if(local_random_state.rand == 0) {
        pthread_mutex_lock(&global_lock);
        thread_number = thread_number + 1;
        pthread_mutex_unlock(&global_lock);
        local_random_state.rand = 1045341691 * thread_number + thread_number *347 + thread_number;
    }
    unsigned int r = local_random_state.rand;
    local_random_state.rand = r * 103515245191 ;
    while(local_random_state.rand == 0) {
        pthread_mutex_lock(&global_lock);
        thread_number = thread_number + 1;
        pthread_mutex_unlock(&global_lock);
        local_random_state.rand = 104534169 * thread_number + thread_number *347 + thread_number;
    }
    return local_random_state.rand;
#else
    uint seed = ((int)(intptr_t)p)*1103515241+ 12345;
    return my_rand_r(&seed);

#endif
}

treap * create_unset_treap_node(){
    return malloc(sizeof(treap));
}

treap * create_treap_node(int key, treap * left, treap * right){
    treap * new_node = create_unset_treap_node();
    new_node->key = key;
    new_node->left = left;
    new_node->right = right;
    new_node->priority = random_int(new_node);
    return new_node;
}

treap * copy(treap * node){
    return create_treap_node(node->key, node->left, node->right);
}

treap * copy_change_left(treap * node, treap * new_left){
    return create_treap_node(node->key, new_left, node->right);
}

treap * copy_change_right(treap * node, treap * new_right){
    return create_treap_node(node->key, node->left, new_right);
}

treap * copy_left_and_right(treap * node, treap * new_left, treap * new_right){
    return create_treap_node(node->key, new_left, new_right);
}


void check_equal_traverser(int v, void* inaux){
    treap * tocheck = (treap *)inaux;
    if(!treap_lookup(tocheck, v)){
      printf("NON EQ IN THREAD %ld value %d\n", pthread_self(), v);
    }
    assert(treap_lookup(tocheck, v));
}

void print_traverser(int v, void* inaux){
    printf("%d ", v);
}

/********* End Help Functions ***************/


treap * treap_create(){
    return NULL;
}

/*
 * input:
 * m = input treap instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * treap = a treap instance containing the items of m + i
 * r_writeback = a treap instance containing the items of m + i
 * r_writeback = false if an item with value i altready existed and false otherwise
 */
treap * treap_insert(treap * m, int i, bool* r_writeback){
    if (m == NULL){
        *r_writeback = true;
        return create_treap_node(i, NULL, NULL);
    } else if (i == m->key){
        *r_writeback = false;
        return copy(m);
    } else if (i < m->key) {
        treap * new_treap = treap_insert(m->left, i, r_writeback);
        if(new_treap->priority > m->priority){
            // Right rotation
            return copy_change_right(new_treap,copy_change_left(m, new_treap->right));
        }else{
            return copy_change_left(m, new_treap);
        }
    } else {
        treap * new_treap = treap_insert(m->right, i, r_writeback);
        if(new_treap->priority > m->priority){
            // Left rotation
            return copy_change_left(new_treap,copy_change_right(m, new_treap->left));
        }else{
            return copy_change_right(m, new_treap);
        }
    }
}

/*
 * input:
 * m = input treap instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * treap = a treap instance containing the items of m + i
 * r_writeback = a treap instance containing the items of m + i
 * r_writeback = false if an item with value i altready existed and false otherwise
 */
treap * treap_remove(treap * m, int i, bool* r_writeback){
    if (m == NULL){
        *r_writeback = false;
        return NULL;
    } else if (i == m->key){
        treap * new_treap;
        if (m->left == NULL && m->right == NULL){
            new_treap = NULL;
        } else if (m->left == NULL){
            new_treap = copy_change_left(m->right,treap_remove(copy_change_right(m, m->right->left), i, r_writeback));
        } else if (m->right == NULL){
            new_treap = copy_change_right(m->left,treap_remove(copy_change_left(m, m->left->right), i, r_writeback));
        } else if(m->left->priority > m->right->priority){
            new_treap = copy_change_right(m->left,treap_remove(copy_change_left(m, m->left->right), i, r_writeback));
        }else{
            new_treap = copy_change_left(m->right,treap_remove(copy_change_right(m, m->right->left), i, r_writeback));
        }
        *r_writeback = true;
        return new_treap;
    } else if (i < m->key) {
        return copy_change_left(m, treap_remove(m->left, i, r_writeback));
    } else {
        return copy_change_right(m, treap_remove(m->right, i, r_writeback));
    }
}


/*
 * input:
 * m = input treap instance
 * i = item to lookup
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
bool treap_lookup(treap * m, int i){
    if (m == NULL){
        return false;
    } else if (i == m->key){
        return true;
    } else if (i < m->key) {
        return treap_lookup(m->left, i);
    } else {
        return treap_lookup(m->right, i);
    }
}


int treap_leftmost_key(treap * m){
    int key;
    do{
        key = m->key;
        m = m->left;
    }while(m != NULL);
    return key;
}

/*
 * input:
 * m = input treap instance
 *
 * output:
 * The key that will be used for splitting
 */
int treap_split_key(treap * m){
    if(m->left != NULL) return m->key;
    else return treap_leftmost_key(m->right);
}


/*
 * input:
 * m = input treap instance
 *
 * output:
 * The left part of the tree when splitting
 */
treap * treap_split_left(treap * m){
    if(m->left != NULL) return m->left;
    else return copy_change_right(m, NULL);
}


/*
 * input:
 * m = input treap instance
 *
 * output:
 * The right part of the tree when splitting
 */
treap * treap_split_right(treap * m){
    bool ignore;
    if(m->left != NULL) return treap_insert(m->right, m->key, &ignore);
    else {
        return m->right;
    }
}

bool treap_is_empty(treap * m){
    return m == NULL;
}

bool treap_max_key(treap * m){
    int key;
    do {
        key = m->key;
        m = m->right;
    } while(m != NULL);
    return key;
}


bool less_than_two_items(treap * m){
    if(m==NULL){
        return true;
    }else if(m->left == NULL && m->right == NULL){
        return true;        
    }else{
        return false;
    }
}


treap * treap_join(treap * t1, treap * t2){
    int remove_key = 10;/* Value does not matter */
    if(t1 == NULL){
        return t2;
    }else if(t2 == NULL){
        return t1;
    }else if(t1->key < t2->key){
        treap * tmp = create_treap_node(remove_key, t1, t2);
        bool ignore;
        return treap_remove(tmp, remove_key, &ignore);
    }else{
        treap * tmp = create_treap_node(remove_key, t2, t1);
        bool ignore;
        return treap_remove(tmp, remove_key, &ignore);
    }
}

/*
 * input:
 * m = input treap instance
 * traverser = a function that will be called once for all items in the range in icreasing order
 *                v = the current item in the range
 *                inaux = will be set to the value passed as the parameter aux
 * aux = some auxallary data (can be used to store items in range)
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
void treap_range_query(treap * m, int start, int end, void (*traverser)(int v, void* inaux), void* aux){
    if (m == NULL){
        return;
    } else if (start < m->key) {
        treap_range_query(m->left, start, end, traverser, aux);
        if(m->key < end){
            traverser(m->key, aux);
            treap_range_query(m->right, start, end, traverser, aux);
        }else if(m->key == end){
            traverser(m->key, aux);
        }         
    } else { // start >= m->key
        if(m->key == start){
            traverser(m->key, aux);
        }
        treap_range_query(m->right, start, end, traverser, aux);
    }
}


void treap_print_range(treap * m, int start, int end){
    printf("RANGE: ");
    treap_range_query(m, start, end, print_traverser, NULL);
}


void treap_check_equal(treap * m1, treap * m2, int start, int end){
    treap_range_query(m1, start, end, check_equal_traverser, m2);
    treap_range_query(m2, start, end, check_equal_traverser, m1);
}

int some_counter;
void treap_print_dot_text_helper(treap * t){
    
    if(t != NULL){
        if(t->left != NULL){
            printf("%d -> %d;\n", t->key, t->left->key);
        }else{
            printf("%d -> NULL%d;\n", t->key, some_counter++);
        }
        if(t->right !=NULL){
            printf("%d -> %d;\n", t->key, t->right->key);
        }else{
            printf("%d -> NULL%d;\n", t->key, some_counter++);
        }
        treap_print_dot_text_helper(t->left);
        treap_print_dot_text_helper(t->right);
    }
}

void treap_print_dot_text(treap * t){
    printf("B(\n");
    printf("digraph G{\n");
    printf("  graph [ordering=\"out\"];\n");
    treap_print_dot_text_helper(t);
    printf("}\n");
    printf("\n)");
}



/* int main (int argc, char *argv[]) { */
/*     unsigned int prev_rand_int = random_int(); */
/*     int upto = 1000000; */
/*     int less = 0; */
/*     for(int i = 0; i < upto; i++){ */
/*         unsigned int rand_int = random_int(); */
/*         //printf("%u %d\n", rand_int, rand_int < prev_rand_int); */
/*         if(rand_int < prev_rand_int){ */
/*             less++; */
/*         } */
/*     } */
/*     printf("less %f\n", ((float)less)/((float)upto)); */
/*     return 0; */
/* } */
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "immutable_treap.h"
#include "random.h"
#include <stdatomic.h>
#include <stdint.h>
#include <assert.h>
#include <limits.h>


#define DEBUG 1
#ifdef DEBUG
#define DEBUG_PRINT(...) do{ printf( __VA_ARGS__ ); } while( false )
#define DEBUG_LFCATREE_PRINT_DOT_TEXT(x) lfcatree_print_dot_text(x);
#define DEBUG_TREAP_PRINT_RANGE(treap, start, end) treap_print_range(treap, start, end)
#else
#define DEBUG_PRINT(...) ;
#define DEBUG_LFCATREE_PRINT_DOT_TEXT(x) ;
#define DEBUG_TREAP_PRINT_RANGE(treap, start, end) ;
#endif
#define NIDHUGG_REMOVE_NONDETERMINISM 1


//#define NIDHUGG_EXPLORE_LESS_UNECESSARY_STUFF 1


#define CONT_CONTRIB 250
#define LOW_CONT_CONTRIB 1
#define RANGE_CONTRIB 100
#define HIGH_CONT 1000
#define LOW_CONT -1000

#define NOT_FOUND ((node*)(uintptr_t)1)

#define NOT_SET ((treap*)(uintptr_t)1)

#define PREPARING ((node*)(uintptr_t)0)
#define DONE ((node*)(uintptr_t)1)
#define ABORTED ((node*)(uintptr_t)2)
#ifdef NIDHUGG_REMOVE_NONDETERMINISM
#define JOINED_NEIGHBOR_STORED ((node*)(uintptr_t)3)
#endif




typedef struct {
    int size;
    int capacity;
    void** stack_array;
    void* first_array[32];
} stack;

stack* new_stack(){
    stack* new_stack = malloc(sizeof(stack));
    new_stack->size = 0;
    new_stack->capacity = 32;
    new_stack->stack_array = new_stack->first_array;
    return new_stack;
}

void push(stack* s, void* item){
    if(s->size == s->capacity){
        void ** prev_array = s->stack_array;
        s->stack_array = malloc(sizeof(void*) * s->capacity * 2);
        for(int i = 0; i < s->capacity; i++){
            s->stack_array[i] = prev_array[i];
        }
        s->capacity = s->capacity * 2;
    }
    s->stack_array[s->size] = item;
    s->size = s->size + 1;
}


void* top(stack* s){
    if(s->size > 0)
        return s->stack_array[s->size - 1];
    else return NULL;
}

void replace_top(stack* s, void* ptr){
    s->stack_array[s->size - 1] = ptr;
}

void* pop(stack* s){
    void* t = top(s);
    if(s->size > 0) s->size = s->size - 1;
    return t;
}

void stack_reset(stack* s){
    s->size = 0;
}

void copy_state_to(stack* from, stack* to){
    to->size = from->size;
    if(to->capacity < from->capacity){
        to->stack_array = malloc(sizeof(void*) * from->capacity);
        to->capacity = from->capacity;
    }
    for(int i = 0; i < from->size; i++){
        to->stack_array[i] = from->stack_array[i];
    }
}


typedef enum {
    route,
    normal,
    join_main,
    join_neighbor,
    range
} node_type;

typedef enum {
    contended = true,
    uncontened = false,
    noinfo = (true + 3)
} contention_info;

struct node_impl;

typedef struct {
    int key;
    volatile atomic_uintptr_t left;
    volatile atomic_uintptr_t right;
    volatile atomic_bool valid;
    volatile atomic_uintptr_t join_id;
} route_node;

typedef struct {
    treap* root;
    int stat;
    struct node_impl* parent;
} normal_base;

typedef struct {
    normal_base normal_data;
    struct node_impl* neigh1;
    struct node_impl* grand_parent;
    struct node_impl* other_branch;
    volatile atomic_uintptr_t joined_neighbor;
#ifdef NIDHUGG_REMOVE_NONDETERMINISM
    struct node_impl* actual_joined_neighbor;
#endif
    treap* linearization_point;
} join_main_base;

typedef struct {
    normal_base normal_data;
    struct node_impl* main_join_node;
    treap* linearization_point;
} join_neighbor_base;


#ifdef NIDHUGG_REMOVE_NONDETERMINISM
#define RESULT_STORAGE_SET ((uintptr_t)2)
#endif

typedef struct {
#ifdef NIDHUGG_REMOVE_NONDETERMINISM
    treap* actual_result;
#endif
  volatile atomic_uintptr_t result;
    volatile atomic_bool more_than_one_base;
    treap* linearization_point;
} result_storage;

typedef result_storage rs;

typedef struct {
    normal_base normal_data;
    int low;
    int high;
    rs* storage;
} range_base;

union node_data {
    route_node route;
    normal_base normal;
    join_main_base join_main;
    join_neighbor_base join_neighbor;
    range_base range;
};

typedef struct node_impl {
    node_type type;
    union node_data data;
} node;

struct lfcatree_impl{
    volatile atomic_uintptr_t root;
    pthread_mutex_t lock;
    treap* model;
};

typedef struct lfcatree_impl lfcat;


#include "lock_free_catree.h"

#define NO_LINEARIZATION_POINT ((treap*)1)

pthread_mutex_t* my_global_lock;


#ifdef NIDHUGG_EXPLORE_LESS_UNECESSARY_STUFF
_Thread_local int global_lock_held_count = 0;

int __VERIFIER_atomic_pthread_mutex_lock(pthread_mutex_t *mutex){
  global_lock_held_count++;
  if(global_lock_held_count > 1){
    return 0;
  }
  return pthread_mutex_lock(mutex);
}

int __VERIFIER_atomic_pthread_mutex_unlock(pthread_mutex_t *mutex){
  global_lock_held_count--;
  if(global_lock_held_count == 0){
    return pthread_mutex_unlock(mutex);
  }else{
    return 0;
  }
}

int my_pthread_mutex_lock(pthread_mutex_t *mutex){
  return __VERIFIER_atomic_pthread_mutex_lock(mutex);
}

int my_pthread_mutex_unlock(pthread_mutex_t *mutex){
  return __VERIFIER_atomic_pthread_mutex_unlock(mutex);
}

#define my_atomic_load(arg, thetype)                            \
  ({                                                            \
    my_pthread_mutex_lock(my_global_lock);                      \
    thetype v = atomic_load(arg);                               \
    my_pthread_mutex_unlock(my_global_lock);                    \
    v;                                                          \
  })

#define my_atomic_store(arg1)                                   \
  do{                                                           \
    my_pthread_mutex_lock(my_global_lock);                      \
    atomic_store(arg1, arg2);                                   \
    my_pthread_mutex_unlock(my_global_lock);                    \
  }while(false)

#define my_atomic_compare_exchange_strong(arg1,arg2,arg3)       \
  do{                                                           \
    my_pthread_mutex_lock(my_global_lock);                      \
    atomic_compare_exchange_strong(arg1,arg2,arg3);             \
    my_pthread_mutex_unlock(my_global_lock);                    \
  }while(false)
  

#else


int my_pthread_mutex_lock(pthread_mutex_t *mutex){
  return pthread_mutex_lock(mutex);
}

int my_pthread_mutex_unlock(pthread_mutex_t *mutex){
  return pthread_mutex_unlock(mutex);
}

#define my_atomic_load(arg, thetype) atomic_load(arg)

#define my_atomic_store(arg1, arg2) atomic_store(arg1, arg2)

#define my_atomic_compare_exchange_strong(arg1,arg2,arg3) atomic_compare_exchange_strong(arg1,arg2,arg3)
#endif


//--------- Help functions ----------------




uintptr_t to_uintptr_t(void* ptr){
  return (uintptr_t)ptr;
}

node* to_nodeptr(uintptr_t ptr){
  return (node*)ptr;
}

void set_linearization_point(node* n, treap* linearization_point){
    if(n->type == join_main){
        n->data.join_main.linearization_point = linearization_point;
    }else{
        n->data.join_neighbor.linearization_point = linearization_point;
    }
}

treap* linearization_point(node* n){
    if(n->type == join_main){
        return n->data.join_main.linearization_point;
    }else if(n->type == join_neighbor){
        return n->data.join_neighbor.linearization_point;
    }else{
        return NO_LINEARIZATION_POINT;
    }
}

treap* data(node* n){
    return n->data.normal.root;
}

node* new_base(node* parent, treap* data, int stat){
    node* new_node = malloc(sizeof(node));
    new_node->type = normal;
    new_node->data.normal.root = data;
    new_node->data.normal.stat = stat;
    new_node->data.normal.parent = parent;
    return new_node;
}

node* new_join_main_base(node* base){
    node* new_node = malloc(sizeof(node));
    new_node->type = join_main;
    new_node->data.join_main.normal_data = base->data.normal;
    new_node->data.join_main.neigh1 = NULL;
    new_node->data.join_main.grand_parent = NULL;
    new_node->data.join_main.other_branch = NULL;
    new_node->data.join_main.joined_neighbor = ATOMIC_VAR_INIT((uintptr_t)PREPARING);
    new_node->data.join_main.linearization_point = NO_LINEARIZATION_POINT;
    return new_node;
}

node* neigh1(node* n){
    return n->data.join_main.neigh1;
}

node* grand_parent(node* n){
    return n->data.join_main.grand_parent;
}

node* other_branch(node* n){
    return n->data.join_main.other_branch;
}

node* new_join_neighbor_base(node* base, node* main_base){
    node* new_node = malloc(sizeof(node));
    new_node->type = join_neighbor;
    new_node->data.join_neighbor.normal_data = base->data.normal;
    new_node->data.join_neighbor.main_join_node = main_base;
    new_node->data.join_neighbor.linearization_point = NO_LINEARIZATION_POINT;
    return new_node;
}

node* new_joined_neighbor(node* base, node* main_base, node* parent){
    node* new_node = malloc(sizeof(node));
    new_node->type = join_neighbor;
    new_node->data.join_neighbor.normal_data = base->data.normal;
    new_node->data.join_neighbor.normal_data.parent = parent;
    new_node->data.join_neighbor.main_join_node = main_base;
    new_node->data.join_neighbor.normal_data.root = treap_join(data(base), data(main_base));
    new_node->data.join_neighbor.linearization_point = NO_LINEARIZATION_POINT;
    return new_node;
}

node*  main_node(node* n){
    return n->data.join_neighbor.main_join_node;
}

rs* new_storage(){
    rs* storage = malloc(sizeof(rs));
    atomic_init( &storage->result, (uintptr_t)NOT_SET );
    atomic_init( &storage->more_than_one_base, false );
    storage->linearization_point = NOT_SET;
    return storage;
}

treap* result(rs* s){
    my_pthread_mutex_lock(my_global_lock);
#ifdef NIDHUGG_REMOVE_NONDETERMINISM
    uintptr_t res = my_atomic_load(&s->result, uintptr_t);
    treap* result;
    if(res == RESULT_STORAGE_SET){
      result = s->actual_result;
    }else{
      result = (treap*)res;
    }
#else
    treap* result = (treap *)my_atomic_load(&s->result, uintptr_t);    
#endif

    my_pthread_mutex_unlock(my_global_lock);
    return result;
}

node* new_range_base(node* base, int low, int high, rs* storage){
    node* new_node = malloc(sizeof(node));
    new_node->type = range;
    new_node->data.range.normal_data = base->data.normal;
    new_node->data.range.low = low;
    new_node->data.range.high = high;
    new_node->data.range.storage = storage;
    return new_node;
}

int low(node* range_base){
    return range_base->data.range.low;
}

int high(node* range_base){
    return range_base->data.range.high;
}

rs* storage(node* range_base){
    return range_base->data.range.storage;
}

int stat(node* n){
    return n->data.normal.stat;
}

bool valid(node* n){
  return my_atomic_load(&n->data.route.valid, bool);
}

volatile atomic_bool* valid_ptrptr(node* n){
    return &n->data.route.valid;
}

volatile atomic_uintptr_t* join_id_ptrptr(node* n){
    return &n->data.route.join_id;
}

node* left(node* n){
  return (node*)my_atomic_load(&n->data.route.left, uintptr_t);
}

node* right(node* n){
  return (node*)my_atomic_load(&n->data.route.right, uintptr_t);
}

volatile atomic_uintptr_t* left_ptrptr(node* n){
    return &n->data.route.left;
}

volatile atomic_uintptr_t* right_ptrptr(node* n){
    return &n->data.route.right;
}

volatile atomic_uintptr_t* joined_neighbor_ptrptr(node* n){
    return &n->data.join_main.joined_neighbor;
}

int key(node* n) {
    return n->data.route.key;
}

int new_stat(node* n, contention_info info);

node* update_base_node(node* org,
                        contention_info cont_info,
                        int value,
                        treap* (*updater)(treap*, int v, bool* res_wb),
                        bool* res){
    treap* new_data = updater(data(org), value, res);
    return new_base(org->data.normal.parent, new_data, new_stat(org, cont_info));
    
}

node* root(lfcatree* m);
/*PICK_OUT_START:additionalcode*/
node* find_base_node(node* n, int i) {
  while(n->type == route){
    if(i < key(n)){
      n = left(n);
    }else{;
      n = right(n);
    }
  }
  return n;
}
node* find_base_stack(node* n, int i, stack* s) {
  stack_reset(s);
  while(n->type == route){
    push(s, n);
    if(i < key(n)){
      n = left(n);
    }else {
      n = right(n);
    }
  }
  push(s, n);
  return n;
}
node* leftmost_and_stack(node* n, stack* s){
  while (n->type == route) {
    push(s, n);
    n = left(n);
  }
  push(s, n);
  return n;    
}
node* parent_of(lfcatree* t, node* n){
  node* prev_node = NULL;
  node* curr_node = root(t);
  while(curr_node != n && curr_node->type == route){
    prev_node = curr_node;
    if(key(n) < key(curr_node)){
      curr_node = left(curr_node);
    }else {
      curr_node = right(curr_node);
    }
  }
  if(curr_node->type != route){
    return NOT_FOUND;
  }
  return prev_node;
}
/*PICK_OUT_END*/

node* find_base_node_parent(node* n, int i) {
  node* prev = NULL;
  while(n->type == route){
    prev = n;
    if(i < key(n)){
      n = left(n);
    }else{
      n = right(n);
    }
  }
  return prev;
}


typedef struct treap_list_impl {
    treap* v;
    node* n;
    struct treap_list_impl* next;
    
} treap_list;

treap_list* treap_list_add(treap_list* l, treap* v, node* n){
    treap_list* new_node = malloc(sizeof(treap_list));
    new_node->v = v;
    new_node->n = n;
    if(l == NULL){
        new_node->next = NULL;
    }else{
        new_node->next = l;
    }
    return new_node;
}


void print_node_type(node* n);
void lfcatree_print_dot_text(lfcatree* t);
node* find_base_node_linearization_point(node* n, int i, lfcatree* thestructure, treap** possible_linearization_point_wb){
    while(n->type == route){
        my_pthread_mutex_unlock(&thestructure->lock);
        if(i < n->data.route.key){
	    my_pthread_mutex_lock(&thestructure->lock);
            n = left(n);
        }else{
	    my_pthread_mutex_lock(&thestructure->lock);
            n = right(n);
        }
    }
    *possible_linearization_point_wb = thestructure->model;
    my_pthread_mutex_unlock(&thestructure->lock);
    return n;
  
}




node* root(lfcatree* m){
  return (node*)my_atomic_load(&m->root, uintptr_t);
}


node* parent(node* n){
    return n->data.normal.parent;
}


bool CAS(volatile atomic_uintptr_t* location, void* expected, void* actual){
    uintptr_t* e = (uintptr_t*)&expected;
    my_pthread_mutex_lock(my_global_lock);
    bool result = atomic_compare_exchange_strong(location,
                                           e, (uintptr_t)actual);
    my_pthread_mutex_unlock(my_global_lock);
    return result;
}

bool CAS_uintptr_t(volatile atomic_uintptr_t* location, uintptr_t expected, uintptr_t actual){
    uintptr_t* e = &expected;
    my_pthread_mutex_lock(my_global_lock);
    bool result = atomic_compare_exchange_strong( location,
                                                  e, (uintptr_t)actual);
    my_pthread_mutex_unlock(my_global_lock);
    return result;
}

bool CAS_bool(volatile atomic_bool* location, bool expected, bool actual){
    bool* e = (bool*)&expected;
    bool result = atomic_compare_exchange_strong( location,
                                                  e, (uintptr_t)actual );
    return result;
}

bool CAS_store_linearization_point(volatile atomic_uintptr_t* location, void* expected, void* actual, node* main_join_node, lfcatree* t){
    uintptr_t* e = (uintptr_t*)&expected;
    my_pthread_mutex_lock(my_global_lock);
    bool result = atomic_compare_exchange_strong( location,
                                                  e, (uintptr_t)actual);
    if(result){
        set_linearization_point(main_join_node, t->model);
    }
    my_pthread_mutex_unlock(my_global_lock);
    return result;
}

bool CAS_store_linearization_points(volatile atomic_uintptr_t* location, void* expected, void* actual, node* main_join_node, node* neighbor_join_node, lfcatree* t){
    uintptr_t* e = (uintptr_t*)&expected;
    my_pthread_mutex_lock(my_global_lock);
    bool result = atomic_compare_exchange_strong( location,
                                                  e, (uintptr_t)actual);
    if(result){
        set_linearization_point(main_join_node, t->model);
        set_linearization_point(neighbor_join_node, t->model);
    }
    my_pthread_mutex_unlock(my_global_lock);
    return result;
}


bool CAS_rs_linearization(volatile atomic_uintptr_t* location,  void* expected,  void* actual, rs* my_storage, lfcatree* t, rs** linearization_rs_wb){
    uintptr_t* e = (uintptr_t*)&expected;
    my_pthread_mutex_lock(my_global_lock);
#ifdef NIDHUGG_REMOVE_NONDETERMINISM
    bool result = atomic_compare_exchange_strong( location,
                                                  e, RESULT_STORAGE_SET);
#else
    bool result = atomic_compare_exchange_strong( location,
                                                  e, (uintptr_t)actual);
#endif
    if(result){
#ifdef NIDHUGG_REMOVE_NONDETERMINISM
      my_storage->actual_result = actual;
#endif
      //        if(*linearization_rs_wb == NULL){
            my_storage->linearization_point = t->model;
            //   }else{
            //     my_storage->linearization_point = (*linearization_rs_wb)->linearization_point;
            //}
    }
    my_pthread_mutex_unlock(my_global_lock);
    return result;
}

void lfcatree_print_dot_text_helper(node* t){
    if(t->type == route){
        node* l = left(t);
        if(l->type == route){
            printf("%d -> %d;\n", key(t), key(l));
        }else{
            printf("%d -> \"(%p) ", key(t), (void*)l);
            treap_print_range(data(l), INT_MIN, INT_MAX);
            printf("\";\n");
        }
        node* r = right(t);
        if(r->type == route){
            printf("%d -> %d;\n", key(t), key(r));
        }else{
            printf("%d -> \"(%p) ", key(t), (void*)r);
            treap_print_range(data(r), INT_MIN, INT_MAX);
            printf("\";\n");
        }
        if(l->type == route) lfcatree_print_dot_text_helper(l);
        if(r->type == route) lfcatree_print_dot_text_helper(r);
    }else{
        print_node_type(t);
        treap_print_range(data(t), INT_MIN, INT_MAX);
        printf("\n");
    }
}

bool CAS_check_with_model(volatile atomic_uintptr_t* location, void* expected, void* actual, lfcatree* thestructure, treap* (*updater)(treap*, int v, bool* res_wb), int i, bool expected_return_value){
    uintptr_t* e = (uintptr_t*)&expected;
    my_pthread_mutex_lock(&thestructure->lock);
    bool success =  atomic_compare_exchange_strong( location,
                                                    e, (uintptr_t)actual);
    if(success){
        //DEBUG_PRINT("UPDATE %d\n", i);
        bool actual_return_value;
        treap* before_change = thestructure->model;
        thestructure->model = updater(thestructure->model, i, &actual_return_value);
        //DEBUG_PRINT("######### SUCCESSFULL CAS ATTEMPT %p -> %p | ", expected, actual);
        //DEBUG_TREAP_PRINT_RANGE(before_change, -10, 10);
        //DEBUG_PRINT(" -> ");
        //DEBUG_TREAP_PRINT_RANGE(thestructure->model, -10, 10);
        //DEBUG_PRINT("\n");
        if(actual_return_value != expected_return_value){
            DEBUG_PRINT("EXPECTED %d (key %d)\n", expected_return_value, i);
            DEBUG_PRINT("expected ");
            DEBUG_TREAP_PRINT_RANGE(data((node*)expected), -10, 10);
            DEBUG_PRINT("\n");
            DEBUG_PRINT("ACTUAL %d\n", actual_return_value);
            DEBUG_PRINT("model before change ");
            DEBUG_TREAP_PRINT_RANGE(before_change, -10, 10);
            DEBUG_PRINT("\n");
        }
        assert(actual_return_value == expected_return_value);
        my_pthread_mutex_unlock(&thestructure->lock);
        return true;
    }else{
        my_pthread_mutex_unlock(&thestructure->lock); 
        return false;
    }
}

/*PICK_OUT_START:data_structures*/
/*REPLACE_START*//*REPLACE_END*/
/*REPLACEMENT_START
// Constants
#define CONT_CONTRIB     250      // Adaption related//(*@\label{code:adaptconstantsstart}@*)
#define LOW_CONT_CONTRIB 1        // ...
#define RANGE_CONTRIB    100      // ...
#define HIGH_CONT        1000     // ...
#define LOW_CONT         -1000    // ...//(*@\label{code:adaptconstantsend}@*)
#define NOT_FOUND        (node*)1 // Special pointers
#define NOT_SET          (treap*)1// ...
#define PREPARING        (node*)0 // Used for join
#define DONE             (node*)1 // ...
#define ABORTED          (node*)2 // ...
enum contention_info { contended, uncontened, noinfo }
// Data Structures
struct route_node {//(*@\label{code:nodetypesstart}@*)//(*@\label{code:routestart}@*)
  int key;                       // Split key//(*@\label{code:keyfield}@*)
  atomic node* left;             // < key
  atomic node* right;            // >= key
  atomic bool valid = true;      // Used for join
  atomic node* join_id = NULL; // ...
}//(*@\label{code:routeend}@*)
struct normal_base {//(*@\label{code:basenodesstart}@*)//(*@\label{code:normalbasenodestart}@*)
  treap* data = NULL;  // Items in the set//(*@\label{code:datafield}@*)
  int stat = 0;        // Statistics variable
  node* parent = NULL; // Parent node or NULL (root) 
}//(*@\label{code:normalbasenodeend}@*)
struct join_main with_fields_from normal_base {//(*@\label{code:complexbasenodesstart}@*)
  node* neigh1;  // First (not joined) neighbor base
  atomic node* neigh2 = PREPARING; // Joined n...
  node* gparent; // Grand parent
  node* otherb;  // Other branch
}
struct join_neighbor with_fields_from normal_base {//(*@\label{code:normalbasenodesstart}@*)
  node* main_node  // The main node for the join
}
struct rs { // Result storage for range queries//(*@\label{code:defresstoragestart}@*)
  atomic treap* result = NOT_SET; // The result
  atomic bool more_than_one_base = false;
}//(*@\label{code:defresstorageend}@*)
struct range_base with_fields_from normal_base {//(*@\label{code:defrangebasestart}@*)
  int lo; int hi; // Low and high key
  rs* storage;
}//(*@\label{code:basenodesend}@*)//(*@\label{code:complexbasenodesend}@*)//(*@\label{code:defrangebaseend}@*)
enum node_type {
  route, normal, join_main, join_neighbor, range
}
struct node with_fields_from normal_base, range_base,
                             join_main, join_neighbor{
  node_type type;
}
struct lfcat{
  atomic node* root;
}//(*@\label{code:nodetypesend}@*)
REPLACEMENT_END*/
// Help functions
bool try_replace(lfcat* m, node* b, node* new_b){//(*@\label{code:tryreplacestart}@*)
  if( parent(b) == NULL )
    return CAS(&m->root, b, new_b);
  else if(left(parent(b)) == b)
    return CAS(left_ptrptr(parent(b)), b, new_b);
  else if(right(parent(b)) == b)
    return CAS(right_ptrptr(parent(b)), b, new_b);
  else return false;
}//(*@\label{code:tryreplaceend}@*)
bool is_replaceable(node* n) {//(*@\label{code:isreplaceblestart}@*)
#ifdef NIDHUGG_REMOVE_NONDETERMINISM/*REMOVE_LINE*/
  uintptr_t v = 0;/*REMOVE_LINE*/
  uintptr_t v2 = 0;/*REMOVE_LINE*/
  if(n->type == join_neighbor){/*REMOVE_LINE*/
    v = my_atomic_load(joined_neighbor_ptrptr(main_node(n)), uintptr_t);/*REMOVE_LINE*/
  }/*REMOVE_LINE*/
  if(n->type == join_main){/*REMOVE_LINE*/
    v2 = my_atomic_load(joined_neighbor_ptrptr(n), uintptr_t);/*REMOVE_LINE*/
  }/*REMOVE_LINE*/
  return (n->type == normal ||/*REMOVE_LINE*/
          (n->type == join_main &&/*REMOVE_LINE*/
           v2 == to_uintptr_t(ABORTED)) ||/*REMOVE_LINE*/
          (n->type == join_neighbor &&/*REMOVE_LINE*/
           (v == to_uintptr_t(ABORTED) ||/*REMOVE_LINE*/
            v == to_uintptr_t(DONE))) ||/*REMOVE_LINE*/
          (n->type == range &&/*REMOVE_LINE*/
           ((treap*)my_atomic_load(&storage(n)->result, uintptr_t)) != NOT_SET));/*REMOVE_LINE*/
#else/*REMOVE_LINE*/
  return (n->type == normal ||
          (n->type == join_main &&
           my_atomic_load(joined_neighbor_ptrptr(n), uintptr_t) == to_uintptr_t(ABORTED)) ||
          (n->type == join_neighbor &&
           (my_atomic_load(joined_neighbor_ptrptr(main_node(n)), uintptr_t) == to_uintptr_t(ABORTED) ||
            my_atomic_load(joined_neighbor_ptrptr(main_node(n)), uintptr_t) == to_uintptr_t(DONE))) ||
          (n->type == range &&
           result(storage(n)) != NOT_SET));
#endif/*REMOVE_LINE*/
}//(*@\label{code:isreplacebleend}@*)
/*PICK_OUT_END*/

// Debug functions


void print_node_type(node* n){
    if(n->type == route) {
        DEBUG_PRINT("route (%p %d)\n", (void*)n, key(n));
    } else if (n->type == normal) {
        DEBUG_PRINT("normal (%p)\n", (void*)n);
    } else if (n->type == join_main) {
      DEBUG_PRINT("join_main joined_neighbor=%p parent=%p  (%p)\n", (void*)my_atomic_load(&n->data.join_main.joined_neighbor, uintptr_t), (void*)parent(n),(void*)n);
    } else if (n->type == join_neighbor) {
      DEBUG_PRINT("join_neighbor (%p) parent=%p status=%p  replacable=%d with main:", (void*)n, (void*)parent(n), (void*)my_atomic_load(&main_node(n)->data.join_main.joined_neighbor, uintptr_t), ( n->type == join_neighbor &&
                                                                                                                                                                                                     ( my_atomic_load(&main_node(n)->data.join_main.joined_neighbor, uintptr_t) == (uintptr_t)ABORTED ||
                                                                                                                                                                                                       my_atomic_load(&main_node(n)->data.join_main.joined_neighbor, uintptr_t) == (uintptr_t)DONE)));
        print_node_type(main_node(n));
    }
}



void lfcatree_print_dot_text(lfcatree* t){
    my_pthread_mutex_lock(my_global_lock);
    DEBUG_PRINT("digraph G{\n");
    DEBUG_PRINT("  graph [ordering=\"out\"];\n");
    lfcatree_print_dot_text_helper(root(t));
    DEBUG_PRINT("}\n");
    my_pthread_mutex_unlock(my_global_lock);
}






bool try_replace(lfcatree* m, node* base, node* new_base);


bool try_replace_check_with_model(lfcatree* m, node* base, node* new_base, lfcatree* thestructure, treap* (*updater)(treap *, int v, bool* res_wb), int i, bool expected_return_value){
    if( parent(base) == NULL ) {
        return CAS_check_with_model(&m->root, base, new_base, thestructure,  updater, i, expected_return_value);
    } else if(left(parent(base)) == base){
        return CAS_check_with_model(left_ptrptr(parent(base)), base, new_base, thestructure,  updater, i, expected_return_value);
    } else if(right(parent(base)) == base){
        return CAS_check_with_model(right_ptrptr(parent(base)), base, new_base, thestructure, updater, i, expected_return_value);
    } else {
        return false;
    }
}


bool is_replaceable(node* n);




node* leftmost_base_node(node* n) {
    while (n->type == route) {
        n = left(n);
    }
    return n;
}

node* rightmost_base_node(node* n) {
    while (n->type == route) {
        n = right(n);
    }
    return n;
}

void set_grand_parent_join_main(node* m, node* gparent){
  m->data.join_main.grand_parent = gparent;
}

void set_other_branch_join_main(node* m, node* other_branch){
  m->data.join_main.other_branch = other_branch;
}

void set_neigh1_join_main(node* m, node* n){
  m->data.join_main.neigh1 = n;
}

node* new_route(node* base, lfcatree* m){
    node* new_route = malloc(sizeof(node));
    new_route->type = route;
    route_node r = {
        .key = treap_split_key(data(base)),
        .left = ATOMIC_VAR_INIT((uintptr_t)new_base(new_route, treap_split_left(data(base)), 0)),
        .right = ATOMIC_VAR_INIT((uintptr_t)new_base(new_route, treap_split_right(data(base)), 0)),
        .valid = ATOMIC_VAR_INIT(true)
    };
    new_route->data.route = r;
    return new_route;
}

node* try_secure_join_right(lfcatree* t, node* b){
  node* n0 = rightmost_base_node(left(parent(b)));//FF
  if( ! is_replaceable(n0) ) return NULL;//FF
  node* m = new_join_main_base(b);//FF
  if( ! CAS(right_ptrptr(parent(b)), b, m) ) return NULL;//FF
  node* n1 = new_join_neighbor_base(n0, m);//FF
  if( ! try_replace(t, n0, n1)) goto fail0;//FF
  if(!CAS(join_id_ptrptr(parent(m)), NULL, m))//FF
    goto fail0;//FF
  node* gparent = parent_of(t, parent(m));//FF
  if(gparent == NOT_FOUND ||//FF
     (gparent != NULL &&//FF
      !CAS(join_id_ptrptr(gparent),NULL,m))) goto fail1;//FF
  set_grand_parent_join_main(m, gparent);//FF
  set_other_branch_join_main(m, left(parent(m)));//FF
  set_neigh1_join_main(m, n1);//FF
  node* joinedp = other_branch(m)==n1 ? gparent: parent(n1);//FF
#ifdef NIDHUGG_REMOVE_NONDETERMINISM/*REPLACE_START*///FF
  m->data.join_main.actual_joined_neighbor =//FF
    new_joined_neighbor(n1, m, joinedp);//FF
  if(CAS(joined_neighbor_ptrptr(m), PREPARING,//FF
         JOINED_NEIGHBOR_STORED))///FF
#else//FF
  if(CAS(joined_neighbor_ptrptr(m), PREPARING,//FF
         new_joined_neighbor(n1, m, joinedp)))//FF
#endif/*REPLACE_END*/
/*REPLACEMENT_START
  if(CAS(joined_neighbor_ptrptr(m), PREPARING,
         new node{... = n, // assign fields from n
                  type = join_neighbor,
                  parent = joinedp,
                  main_join_node = m,
                  data = treap_join(m, n)}))
REPLACEMENT_END*/
    return m;//FF
  if(gparent == NULL) goto fail1;//FF
  atomic_store(join_id_ptrptr(gparent), to_uintptr_t(NULL));//FF
 fail1: atomic_store(join_id_ptrptr(parent(m)), to_uintptr_t(NULL));//FF
 fail0: atomic_store(joined_neighbor_ptrptr(m), to_uintptr_t(ABORTED));//FF
  return NULL;//FF
}


/*PICK_OUT_START:low_contention_adapt*/
node* try_secure_join_left(lfcatree* t, node* b){//(*@\label{code:trysecurejoinleftstart}@*)
  node* n0 = leftmost_base_node(right(parent(b)));//(*@\label{code:trysecurejoinleftfirstline}@*)
  if(!is_replaceable(n0)) return NULL;
  node* m = new_join_main_base(b);/*REPLACE_START*//*REPLACE_END*/
/*REPLACEMENT_START
  node* m = new node{
    ...  = b, // assign fields from b
    type = join_main};
REPLACEMENT_END*/
  if(!CAS(left_ptrptr(parent(b)), b, m)) return NULL;//(*@\label{code:trysecurejoinleftreplaceb}@*)
  node* n1 = new_join_neighbor_base(n0, m);/*REPLACE_START*//*REPLACE_END*/
/*REPLACEMENT_START
  node* n1 = new node{
    ...       = n0, // assign fields from n0
    type      = join_neighbor,
    main_node = m};
REPLACEMENT_END*/
  if(!try_replace(t, n0, n1)) goto fail0;//(*@\label{code:trysecurejoinleftreplacen0}@*)
  if(!CAS(join_id_ptrptr(parent(m)), NULL, m))//(*@\label{code:trysecurejoinleftjoinidstart}@*)
    goto fail0;
  node* gparent = parent_of(t, parent(m));
  if(gparent == NOT_FOUND ||
     (gparent != NULL &&
      !CAS(join_id_ptrptr(gparent),NULL,m))) goto fail1;//(*@\label{code:trysecurejoinleftjoinidend}@*)
  assert(gparent == NULL || valid(gparent));/*REMOVE_LINE*/
  set_grand_parent_join_main(m, gparent);//(*@\label{code:trysecurejoinleftsetinfostart}@*)
  set_other_branch_join_main(m, right(parent(m)));//(*@\label{code:trysecurejoinleftsetotherbranch}@*)
  set_neigh1_join_main(m, n1);//(*@\label{code:trysecurejoinleftsetinfoend}@*)
  node* joinedp = other_branch(m)==n1 ? gparent: parent(n1);//(*@\label{code:trysecurejoinleftfinalstepstart}@*)
#ifdef NIDHUGG_REMOVE_NONDETERMINISM/*REPLACE_START*/
  m->data.join_main.actual_joined_neighbor =
    new_joined_neighbor(n1, m, joinedp);
  if(CAS(joined_neighbor_ptrptr(m), PREPARING,
         JOINED_NEIGHBOR_STORED))
#else
  if(CAS(joined_neighbor_ptrptr(m), PREPARING,
         new_joined_neighbor(n1, m, joinedp)))
#endif/*REPLACE_END*/
/*REPLACEMENT_START
  if(CAS(joined_neighbor_ptrptr(m), PREPARING,
         new node{... = n1, // assign fields from n1
                  type = join_neighbor,
                  parent = joinedp,
                  main_node = m,
                  data = treap_join(m, n1)}))//(*@\label{code:trysecurejoinleftfinalstepend}@*)
REPLACEMENT_END*/
    return m;
  if(gparent == NULL) goto fail1;//(*@\label{code:trysecurejoinleftrollbackstart}@*)
  atomic_store(join_id_ptrptr(gparent), to_uintptr_t(NULL));
 fail1: atomic_store(join_id_ptrptr(parent(m)), to_uintptr_t(NULL));
 fail0: atomic_store(joined_neighbor_ptrptr(m), to_uintptr_t(ABORTED));//(*@\label{code:trysecurejoinleftrollbackend}@*)
  return NULL;
}//(*@\label{code:trysecurejoinleftend}@*)
void complete_join(lfcatree* t, node* m){//(*@\label{code:completejoinstart}@*)
  node* n2 = to_nodeptr(my_atomic_load(joined_neighbor_ptrptr(m), uintptr_t));
#ifdef NIDHUGG_REMOVE_NONDETERMINISM/*REMOVE_LINE*/
  if(n2 == JOINED_NEIGHBOR_STORED)/*REMOVE_LINE*/
    n2 = m->data.join_main.actual_joined_neighbor;/*REMOVE_LINE*/
#endif/*REMOVE_LINE*/
  if(n2 == DONE) return;
  try_replace(t, neigh1(m), n2);//(*@\label{code:completejoinreplacen1}@*)
  atomic_store(valid_ptrptr(parent(m)), false);//(*@\label{code:completejoininvalidateparent}@*)
  node* replacement =//(*@\label{code:completejoindeteminereplacement1}@*)
    other_branch(m) == neigh1(m) ? n2 : other_branch(m);//(*@\label{code:completejoindeteminereplacement2}@*)
  if (grand_parent(m) == NULL){//(*@\label{code:completejoinspliceoutstart}@*)
    if(other_branch(m) == neigh1(m))/*REPLACE_START*/
      CAS_store_linearization_points(&t->root, parent(m), n2, m, n2, t);
    else
      CAS_store_linearization_point(&t->root, parent(m), other_branch(m), m, t);/*REPLACE_END*/
/*REPLACEMENT_START
    CAS(&t->root, parent(m), replacement);
REPLACEMENT_END*/
  }else if(left(grand_parent(m)) == parent(m)){
    if(other_branch(m) == neigh1(m))/*REPLACE_START*/
      CAS_store_linearization_points(left_ptrptr(grand_parent(m)), parent(m), n2, m, n2, t);
    else
      CAS_store_linearization_point(left_ptrptr(grand_parent(m)), parent(m), other_branch(m), m, t);/*REPLACE_END*/
/*REPLACEMENT_START
    CAS(left_ptrptr(grand_parent(m)), parent(m), replacement);
REPLACEMENT_END*/
    CAS(join_id_ptrptr(grand_parent(m)), m, NULL);//(*@\label{code:completejoinresetjoinid}@*)
  }else if(right(grand_parent(m)) == parent(m)){
    if(other_branch(m) == neigh1(m))/*REPLACE_START*/
      CAS_store_linearization_points(right_ptrptr(grand_parent(m)), parent(m), n2, m, n2, t);
    else
      CAS_store_linearization_point(right_ptrptr(grand_parent(m)), parent(m), other_branch(m), m, t);
    CAS(join_id_ptrptr(grand_parent(m)), m, NULL);/*REPLACE_END*/
/*REPLACEMENT_START
    ... // Symmetric case
REPLACEMENT_END*/
  }//(*@\label{code:completejoinspliceoutend}@*)
  atomic_store(joined_neighbor_ptrptr(m), to_uintptr_t(DONE));//(*@\label{code:completejoinmakingdead}@*)
}//(*@\label{code:completejoinend}@*)
void low_contention_adaptation(lfcatree* t, node* b){//(*@\label{code:lowcontadaptstart}@*)
  if(parent(b) == NULL) return;
  if(left(parent(b)) == b){//(*@\label{code:checkleftchildofparent}@*)
    node* m = try_secure_join_left(t, b);
    if (m != NULL) complete_join(t, m);//(*@\label{code:completejoinifsecurejoin}@*)
  }else if (right(parent(b)) == b){
    node* m = try_secure_join_right(t, b);// LL/*REPLACE_START*/
    if(m != NULL) complete_join(t, m);// LL/*REPLACE_END*/
/*REPLACEMENT_START
      ... // Symmetric case
REPLACEMENT_END*/
  }
}//(*@\label{code:lowcontadaptend}@*)
void high_contention_adaptation(lfcatree* m, node* b){//(*@\label{code:highcontadaptstart}@*)
  if(less_than_two_items(data(b))) return;
  node* r = new_route(b, m);/*REPLACE_START*//*REPLACE_END*/
/*REPLACEMENT_START
  node* r = new node{
   type  = route,
   key   = split_key(data(b)),
   left  = new node{type = base, parent = r, stat = 0,
                    data = split_left(data(base))}),
   right = ..., // Symmetric case
   valid = true};
REPLACEMENT_END*/
  try_replace(m, b, r);//(*@\label{code:highcontchange}@*)
}//(*@\label{code:highcontadaptend}@*)
/*PICK_OUT_END*/
void low_contention_adaption_at(lfcatree* t, int key){
  node* b = find_base_node(root(t), key);
  if( ! is_replaceable(b) ) return;
  low_contention_adaptation(t, b);
}


/*PICK_OUT_START:high_contention_adapt*/
//
/*PICK_OUT_END*/


void high_contention_adaption_at(lfcatree* t, int key){
  node* b = find_base_node(root(t), key);
  if( ! is_replaceable(b) ) return;
  high_contention_adaptation(t, b);
}

treap* all_in_range(lfcatree* t, int lo, int hi, rs* help_storage, rs ** linearization_rs_wb);





/*
      o
     / \
    o
   / \
  o   o
 / \
b   o
*/
bool treap_lookup_check(treap* data, int i, treap* possible_linearization_point, node* base, lfcatree* the_stucture){
    bool actual_result = treap_lookup(data, i);
    my_pthread_mutex_lock(&the_stucture->lock);
    if(linearization_point(base) != NO_LINEARIZATION_POINT){
        assert(actual_result == treap_lookup(linearization_point(base), i));
        my_pthread_mutex_unlock(&the_stucture->lock);
        return actual_result;
    }
    my_pthread_mutex_unlock(&the_stucture->lock);
    assert(actual_result == treap_lookup(possible_linearization_point, i));
    return actual_result;
}

//----------------------------------------

pthread_mutexattr_t attr;
lfcatree* lfcatree_create(){
    lfcatree* m = malloc(sizeof(lfcatree));
    m->model = treap_create();
    //pthread_mutexattr_init(&attr);
    //pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE_NP);
    pthread_mutex_init(&m->lock, NULL/*&attr*/);
    my_global_lock = &m->lock;
    atomic_store( &m->root, (uintptr_t)new_base(NULL, treap_create(), 0));
    return m;
}

/*PICK_OUT_START:help_functions*/
// Help functions
void help_if_needed(lfcatree* t, node* n){//(*@\label{code:helpifneededstart}@*)
#ifdef NIDHUGG_REMOVE_NONDETERMINISM/*REMOVE_LINE*/
  if(n->type == join_neighbor) n = main_node(n);/*REMOVE_LINE*/
  uintptr_t v = 0;/*REMOVE_LINE*/
  if(n->type == join_main)/*REMOVE_LINE*/
    v = my_atomic_load(joined_neighbor_ptrptr(n), uintptr_t);/*REMOVE_LINE*/
  if(n->type == join_main &&/*REMOVE_LINE*/
     v == to_uintptr_t(PREPARING)){/*REMOVE_LINE*/
    CAS(joined_neighbor_ptrptr(n), PREPARING, ABORTED);/*REMOVE_LINE*/
  }else if(n->type == join_main &&/*REMOVE_LINE*/
           v > to_uintptr_t(ABORTED)){/*REMOVE_LINE*/
    complete_join(t, n);/*REMOVE_LINE*/
  }else if(n->type == range &&/*REMOVE_LINE*/
           result(storage(n)) == NOT_SET){/*REMOVE_LINE*/
    rs* linearization_rs_throw_away;/*REMOVE_LINE*/
    all_in_range(t, low(n), high(n), storage(n), &linearization_rs_throw_away);/*REMOVE_LINE*/
  }/*REMOVE_LINE*/
#else/*REMOVE_LINE*/
  if(n->type == join_neighbor) n = main_node(n);
  if(n->type == join_main &&
     my_atomic_load(joined_neighbor_ptrptr(n), uintptr_t) == to_uintptr_t(PREPARING)){
    CAS(joined_neighbor_ptrptr(n), PREPARING, ABORTED);//(*@\label{code:killjoin}@*)
  }else if(n->type == join_main &&
           my_atomic_load(joined_neighbor_ptrptr(n), uintptr_t) > to_uintptr_t(ABORTED)){
    complete_join(t, n);//(*@\label{code:helpcompltejoingcall}@*)
  }else if(n->type == range &&
           result(storage(n)) == NOT_SET){
    rs* linearization_rs_throw_away;/*REMOVE_LINE*/
    all_in_range(t, low(n), high(n), storage(n), &linearization_rs_throw_away);
  }
#endif/*REMOVE_LINE*/
}//(*@\label{code:helpifneededend}@*)
int new_stat(node* n, contention_info info){//(*@\label{code:newstatstart}@*)
  int range_sub = 0;
  if(n->type == range &&//(*@\label{code:newstatrangesubstart}@*)
     my_atomic_load(&storage(n)->more_than_one_base, bool))
    range_sub = RANGE_CONTRIB;//(*@\label{code:newstatrangesubend}@*)
  if (info == contended && stat(n) <= HIGH_CONT) {
    return stat(n) + CONT_CONTRIB - range_sub;//(*@\label{code:increasestatvalue}@*)
  }else if(info == uncontened && stat(n) >= LOW_CONT){
    return stat(n) - LOW_CONT_CONTRIB - range_sub;//(*@\label{code:decreasestatvalue}@*)
  }else return stat(n);
}//(*@\label{code:newstatend}@*)
void adapt_if_needed(lfcatree* t, node* b){//(*@\label{code:adaptifneededstart}@*)
  if(!is_replaceable(b)) return;
  else if(new_stat(b, noinfo) > HIGH_CONT)//(*@\label{code:abovehighcontadaptation}@*)
    high_contention_adaptation(t, b);
  else if(new_stat(b, noinfo) < LOW_CONT)//(*@\label{code:belowlowcontadaptation}@*)
    low_contention_adaptation(t, b);
}//(*@\label{code:adaptifneededend}@*)

bool do_update(lfcatree* m,//(*@\label{code:doupdatestart}@*)
               treap*(*u)(treap*,int,bool*), int i){
  contention_info cont_info = uncontened;
  while(true){
    node* base = find_base_node(root(m), i);//(*@\label{code:findbasenodecall}@*)
    if(is_replaceable(base)){//(*@\label{code:doupdateisreplaceablecall}@*)
      bool res;
      treap* new_data = u(data(base), i, &res);/*REPLACE_START*/
      node* newb = new_base(base->data.normal.parent, new_data, new_stat(base, cont_info));/*REPLACE_END*/
/*REPLACEMENT_START
      node* newb = new node{//(*@\label{code:doupdatereplaceattemptstart}@*)
        type = normal,
        parent = base->parent,
        data = u(data(base), i, &res),
        stat = new_stat(base, cont_info)//(*@\label{code:doupdatestatupdate}@*)
      } 
REPLACEMENT_END*/
      if(try_replace_check_with_model(m, base, newb, m, u, i, res)){//(*@\label{code:doupdatereplaceattemptend}@*)
        adapt_if_needed(m, newb);//(*@\label{code:doupdateadaptcall}@*)
        return res;
      }
    }
    cont_info = contended;
    help_if_needed(m, base);//(*@\label{code:doupdatehelpifneededcall}@*)
  }
}//(*@\label{code:doupdateend}@*)
// Public interface
bool lfcatree_insert(lfcat* m, int i){//(*@\label{code:inremstart}@*)
  return do_update(m, treap_insert, i);
}
bool lfcatree_remove(lfcat* m, int i){
  return do_update(m, treap_remove, i);
}//(*@\label{code:inremend}@*)
bool lfcatree_lookup(lfcat* m, int i){//(*@\label{code:lookupstart}@*)
  treap* possible_linarization_point_treap;/*REMOVE_LINE*/
  my_pthread_mutex_lock(&m->lock);/*REMOVE_LINE*/
  node* base = find_base_node_linearization_point(root(m), i, m, &possible_linarization_point_treap);//(*@\label{code:lookupfindbasenode}@*)
  return treap_lookup_check(data(base), i, possible_linarization_point_treap, base, m);//(*@\label{code:lookupactuallookup}@*)
}//(*@\label{code:lookupend}@*)
void lfcatree_range_query(lfcat* m, int lo, int hi,//(*@\label{code:rangequerystart}@*)
           void (*trav)(int, void*), void* aux){
  rs* linearization_rs_throw_away;/*REMOVE_LINE*/
  treap* result = all_in_range(m, lo, hi, NULL, &linearization_rs_throw_away);//(*@\label{code:querycallallinrange}@*)
  treap_range_query(result, lo, hi, trav, aux);//(*@\label{code:querytraversesnapshot}@*)
}//(*@\label{code:rangequeryend}@*)
/*PICK_OUT_END*/



void lf_range_query_test_traverser(int v, void* inaux){
    lfcatree* t = (lfcatree *)inaux;
    assert(lfcatree_lookup(t, v));
}

void lfcatree_lookup_check(lfcatree* m){
    //my_pthread_mutex_lock(&m->lock);
    treap_range_query(m->model, INT_MIN, INT_MAX, lf_range_query_test_traverser, m);
    //my_pthread_mutex_unlock(&m->lock);
}


/*PICK_OUT_START:range_query*/
node* find_next_base_stack(stack* s) {
  node* base = pop(s);
  node* t = top(s);
  if(t == NULL) return NULL;
  if(left(t) == base)
    return leftmost_and_stack(right(t), s);
  int be_greater_than = key(t);
  while(t != NULL){
    if(valid(t) && key(t) > be_greater_than)//(*@\label{code:findnextbaseexplorerightcheck}@*)
      return leftmost_and_stack(right(t), s);
    else { pop(s); t = top(s); }
  }
  return NULL;
}
/*REPLACE_START*//*REPLACE_END*/
/*REPLACEMENT_START
node* new_range_base(node* b, int lo, int hi, rs* s){
  return new node{... = b, // assign fields from b
                  lo = lo, hi = hi, storage = s}; }
REPLACEMENT_END*/
treap*
all_in_range(lfcat* t, int lo, int hi, rs* help_s, rs** linearization_rs_wb){//(*@\label{code:allinrangestart}@*)
  stack* s = new_stack();
  stack* backup_s = new_stack();
  stack* done = new_stack();
  node* b;
  rs* my_s;
 find_first:b = find_base_stack(root(t),lo,s);
  if(help_s != NULL){
    if(b->type != range || help_s != storage(b)){
      *linearization_rs_wb = help_s;/*REMOVE_LINE*/
      return result(help_s);
    }else my_s = help_s;
  }else if(is_replaceable(b)){
    my_s = new_storage();/*REPLACE_START*//*REPLACE_END*/
/*REPLACEMENT_START
    my_s = new rs;
REPLACEMENT_END*/
    node* n = new_range_base(b, lo, hi, my_s);
    if(!try_replace(t, b, n)) goto find_first;
    replace_top(s, n);
  }else if( b->type == range   && high(b) >= hi){
    return all_in_range(t, low(b), high(b), storage(b), linearization_rs_wb);
  }else{
    help_if_needed(t, b);
    goto find_first;
  }
  while(true){   // Find remaining base nodes
    push(done, b);
    copy_state_to(s, backup_s);
    if(!treap_is_empty(data(b)) && treap_max_key(data(b)) >= hi) break;
  find_next_base_node: b = find_next_base_stack(s);
    if(b == NULL) break;
    else if(result(my_s) != NOT_SET){
      *linearization_rs_wb = my_s;/*REMOVE_LINE*/
      return result(my_s);
    }else if(b->type == range && storage(b) == my_s){
      continue;
    }else if(is_replaceable(b)){
      node* n = new_range_base(b, lo, hi, my_s);
      if(try_replace(t, b, n)) {
        replace_top(s, n); continue;
      } else {
        copy_state_to(backup_s, s);
        goto find_next_base_node;
      }
    }else{
      help_if_needed(t, b);
      copy_state_to(backup_s, s);
      goto find_next_base_node;
    }
  }
  treap* res = data(done->stack_array[0]);
  for(int i = 1; i < done->size; i++)
    res = treap_join(res, data(done->stack_array[i]));
  if(CAS_rs_linearization(&my_s->result,NOT_SET,res, my_s, t, linearization_rs_wb) && done->size > 1)//(*@\label{code:rqsetresultstorage}@*)
    atomic_store(&my_s->more_than_one_base, true);
  *linearization_rs_wb = my_s;/*REMOVE_LINE*/
  uint seed = (uint)(intptr_t)my_s/23423;/*REMOVE_LINE*/
  for(int i = 1; i < done->size; i++)/*REPLACE_START*/
    adapt_if_needed(t, done->stack_array[my_rand_r(&seed) % done->size]);/*REPLACE_END*/
/*REPLACEMENT_START
  adapt_if_needed(t, done->array[r() % done->size]);//(*@\label{code:rangeadaptifneeded}@*)
REPLACEMENT_END*/
  return result(my_s);
}//(*@\label{code:allinrangeend}@*)
/*PICK_OUT_END*/

void lfcatree_range_query_check_model(lfcatree* m, int start, int end){
    rs* linearization_point_rs = NULL;
    treap* result = all_in_range(m, start, end, NULL, &linearization_point_rs);
    my_pthread_mutex_lock(my_global_lock); /* Needed to make sure the linearization point is set */
    my_pthread_mutex_unlock(my_global_lock);
    treap_check_equal(result, linearization_point_rs->linearization_point, start, end);
}
#include "padded_types.h"
#include "random.h"
#include "lock_free_catree.h"
#include "test_framework.h"
#include <pthread.h>
#include <stdatomic.h>

#define max(a,b)            (((a) > (b)) ? (a) : (b))

#define min(a,b)            (((a) < (b)) ? (a) : (b))

typedef union {
    unsigned int value;
    char pad[CACHE_LINE_SIZE];
} LLPaddedSeed;

typedef union {
    unsigned long value;
    char pad[CACHE_LINE_SIZE];
} LLPaddedLocalCounter;

typedef struct {
    unsigned long * localInCSCounter;
    unsigned int * localSeed;
} ThreadLocalData;

#ifdef TIME_STOP_CONDITION
LLPaddedBool stop;
#endif 
LLPaddedNonAtomicInt itemsRangeV;
LLPaddedNonAtomicInt numberOfOperationsV; 
LLPaddedNonAtomicDouble getPercentage;
LLPaddedNonAtomicDouble insertPercentage;
LLPaddedNonAtomicDouble deletePercentage;
LLPaddedNonAtomicDouble rangeQueryPercentage;
lfcatree * set;

void  * critical_section_thread(void * threadLocalDataVPtr){
    ThreadLocalData * threadLocalDataPtr = (ThreadLocalData*)threadLocalDataVPtr;
    unsigned long * localInCSCounter = threadLocalDataPtr->localInCSCounter;
    unsigned int * localSeed = threadLocalDataPtr->localSeed;
    int itemsRange = itemsRangeV.value;
    int numberOfOperations = numberOfOperationsV.value;
    double getPercentageV = getPercentage.value;
    double getPlusInsertPercentage = getPercentageV + insertPercentage.value;
    double getPlusInsertPlusDeletePercentage = getPercentageV + insertPercentage.value +
        deletePercentage.value;
#ifdef TIME_STOP_CONDITION
    while(!atomic_load_explicit(&stop.value, memory_order_acquire)){
#else
    for(long i = 0; i < numberOfOperations; i++){
#endif
        //printf("%i\n", i);
        double randomNumber = random_double(localSeed);
        int element = (int)(itemsRange *random_double(localSeed));

        if(randomNumber >= getPlusInsertPlusDeletePercentage){
            int element2 = (int)(itemsRange *random_double(localSeed));
            lfcatree_range_query_check_model(set, min(element, element2), max(element, element2));
            //range query
            //TODO
        }else if(randomNumber >= getPlusInsertPercentage){
            //delete
            //printf("REMOVE %d\n", element);
            lfcatree_remove(set, element);
            //printf("AFTER REMOVE %d\n", element);
        }else if(randomNumber >= getPercentageV){
            //insert
            //printf("INSERT %d\n", element);
            lfcatree_insert(set, element);
            //printf("AFTER INSERT %d\n", element);
        }else {
            //get
            lfcatree_lookup(set, element);
	    //printf("AFTER LOOKUP %d %f\n", element, randomNumber);
        }
        *localInCSCounter = *localInCSCounter + 1;
        //printf("\n");
    }
    return NULL;
}


int test_parallel_ops(int itemsRangeParam,
                      int numberOfOperationsParam,
                      int threadCount,
                      double getPercentageParm,
                      double insertPercentageParm,
                      double deletePercentageParm,
                      double rangeQueryPercentageParm,
                      int rangeQueryRangeSize){
    itemsRangeV.value = itemsRangeParam;
    numberOfOperationsV.value = numberOfOperationsParam;
    insertPercentage.value = insertPercentageParm;
    deletePercentage.value = deletePercentageParm;
    getPercentage.value = getPercentageParm;
    rangeQueryPercentage.value = rangeQueryPercentageParm;
    set = lfcatree_create();
#ifdef TIME_STOP_CONDITION
    struct timespec testTime = {.tv_sec = 5, .tv_nsec = 100000000};
    atomic_store(&stop.value, false);
#endif
    int i = threadCount;
    pthread_t threads[128];
    LLPaddedLocalCounter localInCSCounters[128];
    LLPaddedSeed localSeeds[128];
    ThreadLocalData * threadLocalData = malloc(sizeof(ThreadLocalData) * i);
    for(int n = 0; n < i; n++){
        localInCSCounters[n].value = 0;
        localSeeds[n].value = n;
        threadLocalData[n].localInCSCounter = &localInCSCounters[n].value;
        threadLocalData[n].localSeed = &localSeeds[n].value;
        pthread_create(&threads[n], NULL,
                       critical_section_thread,
                       &threadLocalData[n]);
    }
#ifdef TIME_STOP_CONDITION        
    nanosleep(&testTime, NULL);
    atomic_store(&stop.value, true);
#endif
    unsigned long localInCSCountersSum = 0;
    for(int n = 0; n < i; n++){
        void * resp = NULL;
        pthread_join(threads[n], &resp);
    }
    for(int n = 0; n < i; n++){
        localInCSCountersSum = localInCSCountersSum + 
            localInCSCounters[n].value;
    }
    lfcatree_lookup_check(set);
#ifndef SILENT_TEST
    printf("TOTAL OP COUNT %lu\n", localInCSCountersSum);
#endif
    lfcatree_print_dot_text(set);
    free(threadLocalData);
    return 1;
}

extern pthread_mutex_t global_lock;
int main (int argc, char *argv[]) {
    pthread_mutex_init(&global_lock, NULL);
    if(!(argc == 9 || argc == 1)){
        printf("Wrong number of parameters. Look at the source.\n");
        return 1;
    }

    int itemsInRange;
    int numberOfOperations;
    int threadCount;
    double lookupPercentage;
    double insertPercentage;
    double deletePercentage;
    double rangeQueryPercentage;
    int rangeQueryRangeSize;
    if(argc == 1){
        itemsInRange = 10;
        numberOfOperations = 1;
        threadCount = 5;
        lookupPercentage = 0.2;
        insertPercentage = 0.4;
        deletePercentage = 0.4;
        rangeQueryPercentage = 0.0;
        rangeQueryRangeSize = 0;
    }else{
        int i = 1;
        itemsInRange = atoi(argv[i]);i++;
        numberOfOperations = atoi(argv[i]);i++;
        threadCount = atoi(argv[i]);i++;
        lookupPercentage = atof(argv[i]);i++;
        insertPercentage = atof(argv[i]);i++;
        deletePercentage = atof(argv[i]);i++;
        rangeQueryPercentage = atof(argv[i]);i++;
        rangeQueryRangeSize = atoi(argv[i]);i++;
    }
    
    char str[500];
#ifdef SILENT_TEST
    str[0] = '\n';
#else
    sprintf(str, "test_parallel_ops(\n"
                 "                  itemsInRange         = %d,\n"
#ifndef TIME_STOP_CONDITION
                 "                  numberOfOperations   = %d,\n"
#endif
                 "                  threadCount          = %d,\n"
                 "                  lookupPercentage     = %lf,\n"
                 "                  insertPercentage     = %lf,\n"
                 "                  deletePercentage     = %lf,\n"
                 "                  rangeQueryPercentage = %lf,\n"
                 "                  rangeQueryRangeSize  = %d)",
            itemsInRange,
#ifndef TIME_STOP_CONDITION
            numberOfOperations,
#endif
            threadCount,
            lookupPercentage*100.0,
            insertPercentage*100.0,
            deletePercentage*100.0,
            rangeQueryPercentage*100.0,
            rangeQueryRangeSize);
#endif
   T(test_parallel_ops(itemsInRange,
                        numberOfOperations,
                        threadCount,
                        lookupPercentage,
                        insertPercentage,
                        deletePercentage,
                        rangeQueryPercentage,
                        rangeQueryRangeSize), str);
   /*  T(test_parallel_ops(1000, */
   /*                      200000, */
   /*                      20, */
   /*                      0.2, */
   /*                      0.4, */
   /*                      0.4, */
   /*                      0.0), "test_create()"); */
   /*  T(test_parallel_ops(1000, */
   /*                      200000, */
   /*                      5, */
   /*                      0.2, */
   /*                      0.4, */
   /*                      0.4, */
   /*                      0.0), "test_create()"); */
   /*  T(test_parallel_ops(1000, */
   /*                      200000, */
   /*                      10, */
   /*                      0.2, */
   /*                      0.4, */
   /*                      0.4, */
   /*                      0.0), "test_create()"); */
    
    return 0;
} 
 
