/*
 *  Copyright 2018 Kjell Winblad (kjellwinblad@gmail.com, http://winsh.me)
 *
 *  This file is part of LFCA tree tests.
 *
 *  catrees is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  catrees is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with catrees.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>
#include <stdint.h>
#include "immutable_treap.h"
#include "random.h"
#include <limits.h>
#include <pthread.h>
struct treap_impl {
    int key;
    struct treap_impl * left;
    struct treap_impl * right;
    unsigned int priority;
};

typedef struct treap_impl treap;


/************* Help Structures **************/

struct thread_local_rand_state {
    char pad1[124];
    unsigned int rand;
    char pad2[128];
};


#ifdef USE_THREAD_LOCAL_RANDOM
static __thread struct thread_local_rand_state local_random_state = {.rand = 0};
#endif

static int thread_number = 0;
pthread_mutex_t global_lock = PTHREAD_MUTEX_INITIALIZER;


/************* Help Functions ***************/

/*
 * Creates a random integer that can't be zero
 */
unsigned int random_int(void * p) {
#ifdef NO_RANDOM
  return 42;
#elif defined  USE_THREAD_LOCAL_RANDOM
    if(local_random_state.rand == 0) {
        pthread_mutex_lock(&global_lock);
        thread_number = thread_number + 1;
        pthread_mutex_unlock(&global_lock);
        local_random_state.rand = 1045341691 * thread_number + thread_number *347 + thread_number;
    }
    unsigned int r = local_random_state.rand;
    local_random_state.rand = r * 103515245191 ;
    while(local_random_state.rand == 0) {
        pthread_mutex_lock(&global_lock);
        thread_number = thread_number + 1;
        pthread_mutex_unlock(&global_lock);
        local_random_state.rand = 104534169 * thread_number + thread_number *347 + thread_number;
    }
    return local_random_state.rand;
#else
    uint seed = ((int)(intptr_t)p)*1103515241+ 12345;
    return my_rand_r(&seed);

#endif
}

treap * create_unset_treap_node(){
    return malloc(sizeof(treap));
}

treap * create_treap_node(int key, treap * left, treap * right){
    treap * new_node = create_unset_treap_node();
    new_node->key = key;
    new_node->left = left;
    new_node->right = right;
    new_node->priority = random_int(new_node);
    return new_node;
}

treap * copy(treap * node){
    return create_treap_node(node->key, node->left, node->right);
}

treap * copy_change_left(treap * node, treap * new_left){
    return create_treap_node(node->key, new_left, node->right);
}

treap * copy_change_right(treap * node, treap * new_right){
    return create_treap_node(node->key, node->left, new_right);
}

treap * copy_left_and_right(treap * node, treap * new_left, treap * new_right){
    return create_treap_node(node->key, new_left, new_right);
}


void check_equal_traverser(int v, void* inaux){
    treap * tocheck = (treap *)inaux;
    if(!treap_lookup(tocheck, v)){
      printf("NON EQ IN THREAD %ld value %d\n", pthread_self(), v);
    }
    assert(treap_lookup(tocheck, v));
}

void print_traverser(int v, void* inaux){
    printf("%d ", v);
}

/********* End Help Functions ***************/


treap * treap_create(){
    return NULL;
}

/*
 * input:
 * m = input treap instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * treap = a treap instance containing the items of m + i
 * r_writeback = a treap instance containing the items of m + i
 * r_writeback = false if an item with value i altready existed and false otherwise
 */
treap * treap_insert(treap * m, int i, bool* r_writeback){
    if (m == NULL){
        *r_writeback = true;
        return create_treap_node(i, NULL, NULL);
    } else if (i == m->key){
        *r_writeback = false;
        return copy(m);
    } else if (i < m->key) {
        treap * new_treap = treap_insert(m->left, i, r_writeback);
        if(new_treap->priority > m->priority){
            // Right rotation
            return copy_change_right(new_treap,copy_change_left(m, new_treap->right));
        }else{
            return copy_change_left(m, new_treap);
        }
    } else {
        treap * new_treap = treap_insert(m->right, i, r_writeback);
        if(new_treap->priority > m->priority){
            // Left rotation
            return copy_change_left(new_treap,copy_change_right(m, new_treap->left));
        }else{
            return copy_change_right(m, new_treap);
        }
    }
}

/*
 * input:
 * m = input treap instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * treap = a treap instance containing the items of m + i
 * r_writeback = a treap instance containing the items of m + i
 * r_writeback = false if an item with value i altready existed and false otherwise
 */
treap * treap_remove(treap * m, int i, bool* r_writeback){
    if (m == NULL){
        *r_writeback = false;
        return NULL;
    } else if (i == m->key){
        treap * new_treap;
        if (m->left == NULL && m->right == NULL){
            new_treap = NULL;
        } else if (m->left == NULL){
            new_treap = copy_change_left(m->right,treap_remove(copy_change_right(m, m->right->left), i, r_writeback));
        } else if (m->right == NULL){
            new_treap = copy_change_right(m->left,treap_remove(copy_change_left(m, m->left->right), i, r_writeback));
        } else if(m->left->priority > m->right->priority){
            new_treap = copy_change_right(m->left,treap_remove(copy_change_left(m, m->left->right), i, r_writeback));
        }else{
            new_treap = copy_change_left(m->right,treap_remove(copy_change_right(m, m->right->left), i, r_writeback));
        }
        *r_writeback = true;
        return new_treap;
    } else if (i < m->key) {
        return copy_change_left(m, treap_remove(m->left, i, r_writeback));
    } else {
        return copy_change_right(m, treap_remove(m->right, i, r_writeback));
    }
}


/*
 * input:
 * m = input treap instance
 * i = item to lookup
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
bool treap_lookup(treap * m, int i){
    if (m == NULL){
        return false;
    } else if (i == m->key){
        return true;
    } else if (i < m->key) {
        return treap_lookup(m->left, i);
    } else {
        return treap_lookup(m->right, i);
    }
}


int treap_leftmost_key(treap * m){
    int key;
    do{
        key = m->key;
        m = m->left;
    }while(m != NULL);
    return key;
}

/*
 * input:
 * m = input treap instance
 *
 * output:
 * The key that will be used for splitting
 */
int treap_split_key(treap * m){
    if(m->left != NULL) return m->key;
    else return treap_leftmost_key(m->right);
}


/*
 * input:
 * m = input treap instance
 *
 * output:
 * The left part of the tree when splitting
 */
treap * treap_split_left(treap * m){
    if(m->left != NULL) return m->left;
    else return copy_change_right(m, NULL);
}


/*
 * input:
 * m = input treap instance
 *
 * output:
 * The right part of the tree when splitting
 */
treap * treap_split_right(treap * m){
    bool ignore;
    if(m->left != NULL) return treap_insert(m->right, m->key, &ignore);
    else {
        return m->right;
    }
}

bool treap_is_empty(treap * m){
    return m == NULL;
}

bool treap_max_key(treap * m){
    int key;
    do {
        key = m->key;
        m = m->right;
    } while(m != NULL);
    return key;
}


bool less_than_two_items(treap * m){
    if(m==NULL){
        return true;
    }else if(m->left == NULL && m->right == NULL){
        return true;        
    }else{
        return false;
    }
}


treap * treap_join(treap * t1, treap * t2){
    int remove_key = 10;/* Value does not matter */
    if(t1 == NULL){
        return t2;
    }else if(t2 == NULL){
        return t1;
    }else if(t1->key < t2->key){
        treap * tmp = create_treap_node(remove_key, t1, t2);
        bool ignore;
        return treap_remove(tmp, remove_key, &ignore);
    }else{
        treap * tmp = create_treap_node(remove_key, t2, t1);
        bool ignore;
        return treap_remove(tmp, remove_key, &ignore);
    }
}

/*
 * input:
 * m = input treap instance
 * traverser = a function that will be called once for all items in the range in icreasing order
 *                v = the current item in the range
 *                inaux = will be set to the value passed as the parameter aux
 * aux = some auxallary data (can be used to store items in range)
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
void treap_range_query(treap * m, int start, int end, void (*traverser)(int v, void* inaux), void* aux){
    if (m == NULL){
        return;
    } else if (start < m->key) {
        treap_range_query(m->left, start, end, traverser, aux);
        if(m->key < end){
            traverser(m->key, aux);
            treap_range_query(m->right, start, end, traverser, aux);
        }else if(m->key == end){
            traverser(m->key, aux);
        }         
    } else { // start >= m->key
        if(m->key == start){
            traverser(m->key, aux);
        }
        treap_range_query(m->right, start, end, traverser, aux);
    }
}


void treap_print_range(treap * m, int start, int end){
    printf("RANGE: ");
    treap_range_query(m, start, end, print_traverser, NULL);
}


void treap_check_equal(treap * m1, treap * m2, int start, int end){
    treap_range_query(m1, start, end, check_equal_traverser, m2);
    treap_range_query(m2, start, end, check_equal_traverser, m1);
}

int some_counter;
void treap_print_dot_text_helper(treap * t){
    
    if(t != NULL){
        if(t->left != NULL){
            printf("%d -> %d;\n", t->key, t->left->key);
        }else{
            printf("%d -> NULL%d;\n", t->key, some_counter++);
        }
        if(t->right !=NULL){
            printf("%d -> %d;\n", t->key, t->right->key);
        }else{
            printf("%d -> NULL%d;\n", t->key, some_counter++);
        }
        treap_print_dot_text_helper(t->left);
        treap_print_dot_text_helper(t->right);
    }
}

void treap_print_dot_text(treap * t){
    printf("B(\n");
    printf("digraph G{\n");
    printf("  graph [ordering=\"out\"];\n");
    treap_print_dot_text_helper(t);
    printf("}\n");
    printf("\n)");
}



/* int main (int argc, char *argv[]) { */
/*     unsigned int prev_rand_int = random_int(); */
/*     int upto = 1000000; */
/*     int less = 0; */
/*     for(int i = 0; i < upto; i++){ */
/*         unsigned int rand_int = random_int(); */
/*         //printf("%u %d\n", rand_int, rand_int < prev_rand_int); */
/*         if(rand_int < prev_rand_int){ */
/*             less++; */
/*         } */
/*     } */
/*     printf("less %f\n", ((float)less)/((float)upto)); */
/*     return 0; */
/* } */
