/*
 *  Copyright 2018 Kjell Winblad (kjellwinblad@gmail.com, http://winsh.me)
 *
 *  This file is part of LFCA tree tests.
 *
 *  catrees is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  catrees is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with catrees.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "padded_types.h"
#include "random.h"
#include "lock_free_catree.h"
#include "test_framework.h"
#include <pthread.h>
#include <stdatomic.h>

#define max(a,b)            (((a) > (b)) ? (a) : (b))

#define min(a,b)            (((a) < (b)) ? (a) : (b))

typedef union {
    unsigned int value;
    char pad[CACHE_LINE_SIZE];
} LLPaddedSeed;

typedef union {
    unsigned long value;
    char pad[CACHE_LINE_SIZE];
} LLPaddedLocalCounter;

typedef struct {
    unsigned long * localInCSCounter;
    unsigned int * localSeed;
} ThreadLocalData;

#ifdef TIME_STOP_CONDITION
LLPaddedBool stop;
#endif 
LLPaddedNonAtomicInt itemsRangeV;
LLPaddedNonAtomicInt numberOfOperationsV; 
LLPaddedNonAtomicDouble getPercentage;
LLPaddedNonAtomicDouble insertPercentage;
LLPaddedNonAtomicDouble deletePercentage;
LLPaddedNonAtomicDouble rangeQueryPercentage;
lfcatree * set;

void  * critical_section_thread(void * threadLocalDataVPtr){
    ThreadLocalData * threadLocalDataPtr = (ThreadLocalData*)threadLocalDataVPtr;
    unsigned long * localInCSCounter = threadLocalDataPtr->localInCSCounter;
    unsigned int * localSeed = threadLocalDataPtr->localSeed;
    int itemsRange = itemsRangeV.value;
    int numberOfOperations = numberOfOperationsV.value;
    double getPercentageV = getPercentage.value;
    double getPlusInsertPercentage = getPercentageV + insertPercentage.value;
    double getPlusInsertPlusDeletePercentage = getPercentageV + insertPercentage.value +
        deletePercentage.value;
#ifdef TIME_STOP_CONDITION
    while(!atomic_load_explicit(&stop.value, memory_order_acquire)){
#else
    for(long i = 0; i < numberOfOperations; i++){
#endif
        //printf("%i\n", i);
        double randomNumber = random_double(localSeed);
        int element = (int)(itemsRange *random_double(localSeed));

        if(randomNumber >= getPlusInsertPlusDeletePercentage){
            int element2 = (int)(itemsRange *random_double(localSeed));
            lfcatree_range_query_check_model(set, min(element, element2), max(element, element2));
            //range query
            //TODO
        }else if(randomNumber >= getPlusInsertPercentage){
            //delete
            //printf("REMOVE %d\n", element);
            lfcatree_remove(set, element);
            //printf("AFTER REMOVE %d\n", element);
        }else if(randomNumber >= getPercentageV){
            //insert
            //printf("INSERT %d\n", element);
            lfcatree_insert(set, element);
            //printf("AFTER INSERT %d\n", element);
        }else {
            //get
            lfcatree_lookup(set, element);
	    //printf("AFTER LOOKUP %d %f\n", element, randomNumber);
        }
        *localInCSCounter = *localInCSCounter + 1;
        //printf("\n");
    }
    return NULL;
}


int test_parallel_ops(int itemsRangeParam,
                      int numberOfOperationsParam,
                      int threadCount,
                      double getPercentageParm,
                      double insertPercentageParm,
                      double deletePercentageParm,
                      double rangeQueryPercentageParm,
                      int rangeQueryRangeSize){
    itemsRangeV.value = itemsRangeParam;
    numberOfOperationsV.value = numberOfOperationsParam;
    insertPercentage.value = insertPercentageParm;
    deletePercentage.value = deletePercentageParm;
    getPercentage.value = getPercentageParm;
    rangeQueryPercentage.value = rangeQueryPercentageParm;
    set = lfcatree_create();
#ifdef TIME_STOP_CONDITION
    struct timespec testTime = {.tv_sec = 5, .tv_nsec = 100000000};
    atomic_store(&stop.value, false);
#endif
    int i = threadCount;
    pthread_t threads[128];
    LLPaddedLocalCounter localInCSCounters[128];
    LLPaddedSeed localSeeds[128];
    ThreadLocalData * threadLocalData = malloc(sizeof(ThreadLocalData) * i);
    for(int n = 0; n < i; n++){
        localInCSCounters[n].value = 0;
        localSeeds[n].value = n;
        threadLocalData[n].localInCSCounter = &localInCSCounters[n].value;
        threadLocalData[n].localSeed = &localSeeds[n].value;
        pthread_create(&threads[n], NULL,
                       critical_section_thread,
                       &threadLocalData[n]);
    }
#ifdef TIME_STOP_CONDITION        
    nanosleep(&testTime, NULL);
    atomic_store(&stop.value, true);
#endif
    unsigned long localInCSCountersSum = 0;
    for(int n = 0; n < i; n++){
        void * resp = NULL;
        pthread_join(threads[n], &resp);
    }
    for(int n = 0; n < i; n++){
        localInCSCountersSum = localInCSCountersSum + 
            localInCSCounters[n].value;
    }
    lfcatree_lookup_check(set);
#ifndef SILENT_TEST
    printf("TOTAL OP COUNT %lu\n", localInCSCountersSum);
#endif
    lfcatree_print_dot_text(set);
    free(threadLocalData);
    return 1;
}

extern pthread_mutex_t global_lock;
int main (int argc, char *argv[]) {
    pthread_mutex_init(&global_lock, NULL);
    if(!(argc == 9 || argc == 1)){
        printf("Wrong number of parameters. Look at the source.\n");
        return 1;
    }

    int itemsInRange;
    int numberOfOperations;
    int threadCount;
    double lookupPercentage;
    double insertPercentage;
    double deletePercentage;
    double rangeQueryPercentage;
    int rangeQueryRangeSize;
    if(argc == 1){
        itemsInRange = 10;
        numberOfOperations = 1;
        threadCount = 5;
        lookupPercentage = 0.2;
        insertPercentage = 0.4;
        deletePercentage = 0.4;
        rangeQueryPercentage = 0.0;
        rangeQueryRangeSize = 0;
    }else{
        int i = 1;
        itemsInRange = atoi(argv[i]);i++;
        numberOfOperations = atoi(argv[i]);i++;
        threadCount = atoi(argv[i]);i++;
        lookupPercentage = atof(argv[i]);i++;
        insertPercentage = atof(argv[i]);i++;
        deletePercentage = atof(argv[i]);i++;
        rangeQueryPercentage = atof(argv[i]);i++;
        rangeQueryRangeSize = atoi(argv[i]);i++;
    }
    
    char str[500];
#ifdef SILENT_TEST
    str[0] = '\n';
#else
    sprintf(str, "test_parallel_ops(\n"
                 "                  itemsInRange         = %d,\n"
#ifndef TIME_STOP_CONDITION
                 "                  numberOfOperations   = %d,\n"
#endif
                 "                  threadCount          = %d,\n"
                 "                  lookupPercentage     = %lf,\n"
                 "                  insertPercentage     = %lf,\n"
                 "                  deletePercentage     = %lf,\n"
                 "                  rangeQueryPercentage = %lf,\n"
                 "                  rangeQueryRangeSize  = %d)",
            itemsInRange,
#ifndef TIME_STOP_CONDITION
            numberOfOperations,
#endif
            threadCount,
            lookupPercentage*100.0,
            insertPercentage*100.0,
            deletePercentage*100.0,
            rangeQueryPercentage*100.0,
            rangeQueryRangeSize);
#endif
   T(test_parallel_ops(itemsInRange,
                        numberOfOperations,
                        threadCount,
                        lookupPercentage,
                        insertPercentage,
                        deletePercentage,
                        rangeQueryPercentage,
                        rangeQueryRangeSize), str);
   /*  T(test_parallel_ops(1000, */
   /*                      200000, */
   /*                      20, */
   /*                      0.2, */
   /*                      0.4, */
   /*                      0.4, */
   /*                      0.0), "test_create()"); */
   /*  T(test_parallel_ops(1000, */
   /*                      200000, */
   /*                      5, */
   /*                      0.2, */
   /*                      0.4, */
   /*                      0.4, */
   /*                      0.0), "test_create()"); */
   /*  T(test_parallel_ops(1000, */
   /*                      200000, */
   /*                      10, */
   /*                      0.2, */
   /*                      0.4, */
   /*                      0.4, */
   /*                      0.0), "test_create()"); */
    
    return 0;
} 
 
